# -*- coding: utf-8 -*-
import sys

import kiwipy

# pylint: disable=invalid-name

message = ' '.join(sys.argv[1:]) or 'Hello World!'

with kiwipy.connect('amqp://127.0.0.1') as comm:
    result = comm.task_send(message).result(timeout=5.0)
    print(result)



# -*- coding: utf-8 -*-
import asyncio
from functools import partial
import copy
import logging
import typing
from typing import Union, Optional, Dict

import shortuuid
import aio_pika

import kiwipy
from . import defaults
from . import tasks
from . import messages
from . import utils

__all__ = 'RmqCommunicator', 'async_connect'

_LOGGER = logging.getLogger(__name__)

# The exchange properties use by the publisher and subscriber.  These have to match
# which is why they're declare her
EXCHANGE_PROPERTIES = {'type': aio_pika.ExchangeType.TOPIC}


class RmqPublisher(messages.BasePublisherWithReplyQueue):
    """
    Publisher for sending a range of message types over RMQ
    """
    DEFAULT_EXCHANGE_PARAMS = EXCHANGE_PROPERTIES

    async def rpc_send(self, recipient_id, msg):
        routing_key = f'{defaults.RPC_TOPIC}.{recipient_id}'
        _LOGGER.debug(
            'Sending RPC with routing key %r to RMQ queue %r: %r',
            routing_key,
            self._reply_queue.name,
            msg,
        )
        message = aio_pika.Message(body=self._encode(msg), reply_to=self._reply_queue.name)
        published, response_future = await self.publish_expect_response(
            message, routing_key=routing_key, mandatory=True
        )
        assert published, 'The message was not published to the exchanges'
        return response_future

    async def broadcast_send(self, msg, sender=None, subject=None, correlation_id=None):
        message_dict = messages.BroadcastMessage.create(
            body=msg,
            sender=sender,
            subject=subject,
            correlation_id=correlation_id,
        )
        _LOGGER.debug(
            'Sending broadcast with routing key %r to RMQ via exchange %r: %r',
            defaults.BROADCAST_TOPIC,
            self._exchange_name,
            message_dict,
        )
        message = aio_pika.Message(
            body=self._encode(message_dict),
            delivery_mode=aio_pika.DeliveryMode.NOT_PERSISTENT,
        )
        # Send as mandatory=False because we don't expect the message to be routable to anyone
        return await self.publish(message, routing_key=defaults.BROADCAST_TOPIC, mandatory=False)


class RmqSubscriber:
    """
    Subscriber for receiving a range of messages over RMQ
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        connection,
        message_exchange=defaults.MESSAGE_EXCHANGE,
        queue_expires=defaults.QUEUE_EXPIRES,
        decoder=defaults.DECODER,
        encoder=defaults.ENCODER,
        testing_mode=False
    ):
        # pylint: disable=too-many-arguments
        """
        Subscribes and listens for process control messages and acts on them
        by calling the corresponding methods of the process manager.

        :param connection: The tokpia connection
        :type connection: :class:`aio_pika.Connection`
        :param message_exchange: The name of the exchange to use
        :param queue_expires: the expiry time for standard queues in milliseconds. This is the time after which, if
            there are no subscribers, a queue will automatically be deleted by RabbitMQ.
        :type queue_expires: int
        :param encoder: The encoder to call for encoding a message
        :param decoder: The decoder to call for decoding a message
        :param testing_mode: Run in testing mode: all queues and exchanges will be temporary
        """
        super().__init__()

        self._connection = connection
        self._channel = None  # type: typing.Optional[aio_pika.Channel]
        self._exchange = None  # type: typing.Optional[aio_pika.Exchange]
        self._exchange_name = message_exchange
        self._decode = decoder
        self._testing_mode = testing_mode
        self._response_encode = encoder

        self._broadcast_queue_arguments = {'x-message-ttl': defaults.MESSAGE_TTL}

        self._rmq_queue_arguments = {'x-message-ttl': defaults.MESSAGE_TTL}
        if queue_expires:
            self._rmq_queue_arguments['x-expires'] = queue_expires

        self._rpc_subscribers = {}
        self._broadcast_subscribers = {}
        self._broadcast_queue = None  # type: typing.Optional[aio_pika.Queue]
        self._broadcast_consumer_tag = None

    async def add_rpc_subscriber(self, subscriber, identifier=None):
        # Create an RPC queue
        rpc_queue = await self._channel.declare_queue(exclusive=True, arguments=self._rmq_queue_arguments)
        try:
            identifier = await rpc_queue.consume(partial(self._on_rpc, subscriber), consumer_tag=identifier)
        except aio_pika.exceptions.DuplicateConsumerTag:
            raise kiwipy.DuplicateSubscriberIdentifier(f"RPC identifier '{identifier}'")
        else:
            await rpc_queue.bind(self._exchange, routing_key=f'{defaults.RPC_TOPIC}.{identifier}')
            # Save the queue so we can cancel and unbind later
            self._rpc_subscribers[identifier] = rpc_queue
            return identifier

    async def remove_rpc_subscriber(self, identifier):
        try:
            rpc_queue = self._rpc_subscribers.pop(identifier)
        except KeyError:
            raise ValueError(f"Unknown subscriber '{identifier}'")
        else:
            await rpc_queue.cancel(identifier)
            await rpc_queue.unbind(self._exchange, routing_key=f'{defaults.RPC_TOPIC}.{identifier}')

    async def add_broadcast_subscriber(self, subscriber, identifier=None):
        identifier = identifier or shortuuid.uuid()
        if identifier in self._broadcast_subscribers:
            raise kiwipy.DuplicateSubscriberIdentifier(f"Broadcast identifier '{identifier}'")

        self._broadcast_subscribers[identifier] = subscriber
        if self._broadcast_consumer_tag is None:
            # Consume on the broadcast queue
            self._broadcast_consumer_tag = await self._broadcast_queue.consume(self._on_broadcast)
        return identifier

    async def remove_broadcast_subscriber(self, identifier):
        try:
            del self._broadcast_subscribers[identifier]
        except KeyError:
            raise ValueError(f"Broadcast subscriber '{identifier}' unknown")
        if not self._broadcast_subscribers:
            await self._broadcast_queue.cancel(self._broadcast_consumer_tag)
            self._broadcast_consumer_tag = None

    def channel(self):
        return self._channel

    async def connect(self):
        """Get a channel and set up all the exchanges/queues we need"""
        if self._channel:
            # Already connected
            return

        exchange_params = copy.copy(EXCHANGE_PROPERTIES)

        if self._testing_mode:
            exchange_params.setdefault('auto_delete', self._testing_mode)

        self._channel = await self._connection.channel()
        self._exchange = await self._channel.declare_exchange(name=self._exchange_name, **exchange_params)

        await self._create_broadcast_queue()

    async def _create_broadcast_queue(self):
        """
        Create and bind the broadcast queue

        One is used for all broadcasts on this exchange
        """
        # Create a new such that we can see this is the broadcast queue
        name = f'broadcast-{shortuuid.uuid()}'
        self._broadcast_queue = await self._channel.declare_queue(
            name=name, exclusive=True, arguments=self._broadcast_queue_arguments
        )
        await self._broadcast_queue.bind(self._exchange, routing_key=defaults.BROADCAST_TOPIC)

    async def disconnect(self):
        await self._channel.close()
        self._exchange = None
        self._channel = None

    async def _on_rpc(self, subscriber, message):
        """
        :param subscriber: the subscriber function or coroutine that will get the RPC message
        :param message: the RMQ message
        :type message: :class:`aio_pika.IncomingMessage`
        """
        async with message.process(ignore_processed=True):
            # Tell the sender that we've dealt with it
            message.ack()
            msg = self._decode(message.body)

            try:
                receiver = utils.ensure_coroutine(subscriber)
                result = await receiver(self, msg)
            except Exception as exc:  # pylint: disable=broad-except
                # We had an exception in  calling the receiver
                await self._send_response(message.reply_to, message.correlation_id, utils.exception_response(exc))
            else:
                if asyncio.isfuture(result):
                    await self._send_future_response(result, message.reply_to, message.correlation_id)
                else:
                    # All good, send the response out
                    await self._send_response(message.reply_to, message.correlation_id, utils.result_response(result))

    async def _on_broadcast(self, message):
        async with message.process():
            msg = self._decode(message.body)
            for receiver in self._broadcast_subscribers.values():
                try:
                    receiver = utils.ensure_coroutine(receiver)
                    await receiver(
                        self, msg[messages.BroadcastMessage.BODY], msg[messages.BroadcastMessage.SENDER],
                        msg[messages.BroadcastMessage.SUBJECT], msg[messages.BroadcastMessage.CORRELATION_ID]
                    )
                except Exception:  # pylint: disable=broad-except
                    _LOGGER.exception('Exception in broadcast receiver')

    async def _send_future_response(self, future, reply_to, correlation_id):
        """
        The RPC call returned a future which means we need to send a pending response
        and send a further message when the future resolves.  If it resolves to another future
        we should send out a further pending response and so on.

        :param future: the future from the RPC call
        :type future: :class:`asyncio.Future`
        :param reply_to: the recipient
        :param correlation_id: the correlation id
        """
        try:
            # Keep looping in case we're in a situation where a future resolves to a future etc.
            while asyncio.isfuture(future):
                # Send out a message saying that we're waiting for a future to complete
                await self._send_response(reply_to, correlation_id, utils.pending_response())
                future = await future
        except kiwipy.CancelledError as exc:
            # Send out a cancelled response
            await self._send_response(reply_to, correlation_id, utils.cancelled_response(str(exc)))
        except Exception as exc:  # pylint: disable=broad-except
            # Send out an exception response
            await self._send_response(reply_to, correlation_id, utils.exception_response(exc))
        else:
            # We have a final result so send that as the response
            await self._send_response(reply_to, correlation_id, utils.result_response(future))

    async def _send_response(self, reply_to, correlation_id, response):
        assert reply_to, 'Must provide an identifier for the recipient'

        message = aio_pika.Message(body=self._response_encode(response), correlation_id=correlation_id)
        result = await self._exchange.publish(message, routing_key=reply_to)
        return result


class RmqCommunicator:
    """
    An asynchronous communicator that relies on aio_pika to make a connection to a RabbitMQ server
    and uses an asyncio event loop for scheduling coroutines and callbacks.
    """

    # pylint: disable=too-many-instance-attributes

    _connection = None
    _message_subscriber = None
    _message_publisher = None
    _default_task_queue = None  # type: Optional[tasks.RmqTaskQueue]

    def __init__(
        self,
        connection: aio_pika.Connection,
        # Messages
        message_exchange: str = defaults.MESSAGE_EXCHANGE,
        queue_expires: int = defaults.QUEUE_EXPIRES,
        # Tasks
        task_exchange: str = defaults.TASK_EXCHANGE,
        task_queue: str = defaults.TASK_QUEUE,
        task_prefetch_size=defaults.TASK_PREFETCH_SIZE,
        task_prefetch_count=defaults.TASK_PREFETCH_COUNT,
        encoder=defaults.ENCODER,
        decoder=defaults.DECODER,
        testing_mode=False
    ):
        # pylint: disable=too-many-arguments
        """Create a new asynchronous communicator.

        .. note: this communicator takes ownership of the connection and, therefore, it should not be shared as when
            this communicator disconnects it will also hang up the connection.

        :param connection: An aio_pika connection, doesn't need to be connected
        :param message_exchange: The name of the RMQ message exchange to use
        :param queue_expires: the expiry time for standard queues in milliseconds.  This is the time after which, if
            there are no subscribers, a queue will automatically be deleted by RabbitMQ.
        :param task_exchange: The name of the RMQ task exchange to use
        :param task_queue: The name of the task queue to use
        :param task_prefetch_count: the number of tasks this communicator can fetch simultaneously
        :param task_prefetch_size: the total size of the messages that the default queue can fetch simultaneously
        :param encoder: The encoder to call for encoding a message
        :param decoder: The decoder to call for decoding a message
        :param testing_mode: Run in testing mode: all queues and exchanges will be temporary
        """
        super().__init__()

        self._connection = connection
        self._loop = connection.loop

        # Save some of these settings for later
        self._message_exchange = message_exchange
        self._queue_expires = queue_expires

        # Default tasks queue
        self._task_exchange = task_exchange
        self._task_queue = task_queue
        self._task_prefetch_size = task_prefetch_size
        self._task_prefetch_count = task_prefetch_count
        self._task_queues = []

        self._decoder = decoder
        self._encoder = encoder
        self._testing_mode = testing_mode

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.disconnect()

    def __str__(self):
        return f'RMQCommunicator({self._connection})'

    @property
    def server_properties(self) -> Dict:
        """
        A dictionary containing server properties as returned by the RMQ server at connection time.

        The details are defined by the RMQ standard and can be found here:

        https://www.rabbitmq.com/amqp-0-9-1-reference.html#connection.start.server-properties

        The protocol states that this dictionary SHOULD contain at least:
            'host'          - specifying the server host name or address
            'product'       - giving the name of the server product
            'version'       - giving the name of the server version
            'platform'      - giving the name of the operating system
            'copyright'     - if appropriate, and,
            'information'   - giving other general information

        .. note:: In testing it seems like 'host' is not always returned.  Host information may be found in
            'cluster_name' but clients shouldn't rely on this.

        :return: the server properties dictionary
        """
        if self._connection is None:
            return {}

        return self._connection.connection.server_properties

    @property
    def loop(self):
        """Get the event loop instance driving this communicator connection."""
        return self._connection.loop

    def add_close_callback(self, callback: aio_pika.types.CloseCallbackType, weak: bool = False) -> None:
        """Add a callable to be called each time (after) the connection is closed.

        :param weak: If True, the callback will be added to a `WeakSet`
        """
        self._connection.add_close_callback(callback, weak)

    async def get_default_task_queue(self) -> tasks.RmqTaskQueue:
        """Get a default task queue.

        If one doesn't exist it will be created as part of this call.
        """
        self._ensure_connected()

        if self._default_task_queue is None:
            task_queue = tasks.RmqTaskQueue(
                self._connection,
                exchange_name=self._task_exchange,
                queue_name=self._task_queue,
                decoder=self._decoder,
                encoder=self._encoder,
                prefetch_size=self._task_prefetch_size,
                prefetch_count=self._task_prefetch_count,
                testing_mode=self._testing_mode
            )

            await task_queue.connect()
            self._default_task_queue = task_queue

        return self._default_task_queue

    async def get_message_subscriber(self) -> RmqSubscriber:
        """Get the message subscriber.

        If one doesn't exist it will be created as part of this call.
        """
        self._ensure_connected()

        if self._message_subscriber is None:
            subscriber = RmqSubscriber(
                self._connection,
                message_exchange=self._message_exchange,
                queue_expires=self._queue_expires,
                encoder=self._encoder,
                decoder=self._decoder,
                testing_mode=self._testing_mode
            )
            await subscriber.connect()
            self._message_subscriber = subscriber

        return self._message_subscriber

    async def get_message_publisher(self) -> RmqPublisher:
        """Get a message publisher.

        If one doesn't exist it will be created as part of this call.
        """
        self._ensure_connected()

        if self._message_publisher is None:
            publisher = RmqPublisher(
                self._connection,
                exchange_name=self._message_exchange,
                encoder=self._encoder,
                decoder=self._decoder,
                testing_mode=self._testing_mode
            )

            await publisher.connect()
            self._message_publisher = publisher

        return self._message_publisher

    def connected(self) -> bool:
        return self._connection is not None and not self._connection.is_closed

    async def connect(self):
        """Establish a connection if not already connected."""
        if not self.connected():
            await self._connection.connect()

    async def disconnect(self):
        """Disconnect from the connection if connected."""
        if not self.connected():
            return

        if self._message_publisher is not None:
            await self._message_publisher.disconnect()
            self._message_publisher = None

        if self._message_subscriber is not None:
            await self._message_subscriber.disconnect()
            self._message_subscriber = None

        if self._default_task_queue is not None:
            await self._default_task_queue.disconnect()
            self._default_task_queue = None

        await self._connection.close()

    async def add_rpc_subscriber(self, subscriber, identifier=None):
        msg_subscriber = await self.get_message_subscriber()
        identifier = await msg_subscriber.add_rpc_subscriber(subscriber, identifier)
        return identifier

    async def remove_rpc_subscriber(self, identifier):
        msg_subscriber = await self.get_message_subscriber()
        await msg_subscriber.remove_rpc_subscriber(identifier)

    async def add_task_subscriber(self, subscriber, identifier=None):
        default_task_queue = await self.get_default_task_queue()
        return await default_task_queue.add_task_subscriber(subscriber, identifier)

    async def remove_task_subscriber(self, identifier):
        default_task_queue = await self.get_default_task_queue()
        await default_task_queue.remove_task_subscriber(identifier)

    async def add_broadcast_subscriber(self, subscriber, identifier=None):
        msg_subscriber = await self.get_message_subscriber()
        identifier = await msg_subscriber.add_broadcast_subscriber(subscriber, identifier)
        return identifier

    async def remove_broadcast_subscriber(self, identifier):
        msg_subscriber = await self.get_message_subscriber()
        await msg_subscriber.remove_broadcast_subscriber(identifier)

    async def rpc_send(self, recipient_id, msg):
        """Initiate a remote procedure call on a recipient.

        :param recipient_id: The recipient identifier
        :param msg: The body of the message
        :return: A future corresponding to the outcome of the call
        """
        try:
            publisher = await self.get_message_publisher()
            response_future = await publisher.rpc_send(recipient_id, msg)
            return response_future
        except aio_pika.exceptions.DeliveryError as exception:
            raise kiwipy.UnroutableError(str(exception))

    async def broadcast_send(self, body, sender=None, subject=None, correlation_id=None):
        publisher = await self.get_message_publisher()
        result = await publisher.broadcast_send(body, sender, subject, correlation_id)
        return result

    async def task_send(self, task, no_reply=False):
        try:
            task_queue = await self.get_default_task_queue()
            result = await task_queue.task_send(task, no_reply)
            return result
        except aio_pika.exceptions.DeliveryError as exception:
            raise kiwipy.UnroutableError(str(exception))
        except aio_pika.exceptions.AMQPError as exception:
            # Find out what the exception is when a nack is generated!
            raise kiwipy.TaskRejected(str(exception))

    async def task_queue(
        self,
        queue_name: str,
        prefetch_size=defaults.TASK_PREFETCH_SIZE,
        prefetch_count=defaults.TASK_PREFETCH_COUNT
    ) -> tasks.RmqTaskQueue:
        """Create a new task queue."""
        queue = tasks.RmqTaskQueue(
            self._connection,
            exchange_name=self._task_exchange,
            queue_name=queue_name,
            decoder=self._decoder,
            encoder=self._encoder,
            prefetch_size=prefetch_size,
            prefetch_count=prefetch_count,
            testing_mode=self._testing_mode
        )
        await queue.connect()
        self._task_queues.append(queue)
        return queue

    def _ensure_connected(self):
        if not self.connected():
            raise RuntimeError(
                'The communicator is not connected, call connect() or use in a context to establish a connection.'
            )


async def async_connect(
    # Connection parameters
    connection_params: Union[str, dict] = None,
    connection_factory=aio_pika.connect_robust,
    # Messages
    message_exchange: str = defaults.MESSAGE_EXCHANGE,
    queue_expires: int = defaults.QUEUE_EXPIRES,
    # Tasks
    task_exchange: str = defaults.TASK_EXCHANGE,
    task_queue: str = defaults.TASK_QUEUE,
    task_prefetch_size=defaults.TASK_PREFETCH_SIZE,
    task_prefetch_count=defaults.TASK_PREFETCH_COUNT,
    encoder=defaults.ENCODER,
    decoder=defaults.DECODER,
    testing_mode=False,
) -> RmqCommunicator:
    # pylint: disable=too-many-arguments
    """Convenience method that returns a connected communicator.

    :param connection_params: parameters that will be passed to the connection factory to create the connection
    :param connection_factory: the factory method to open the aio_pika connection with
    :param message_exchange: The name of the RMQ message exchange to use
    :param queue_expires: the expiry time for standard queues in milliseconds. This is the time after which, if there
        are no subscribers, a queue will automatically be deleted by RabbitMQ.
    :param task_exchange: The name of the RMQ task exchange to use
    :param task_queue: The name of the task queue to use
    :param task_prefetch_count: the number of tasks this communicator can fetch simultaneously
    :param task_prefetch_size: the total size of the messages that the default queue can fetch simultaneously
    :param encoder: The encoder to call for encoding a message
    :param decoder: The decoder to call for decoding a message
    :param testing_mode: Run in testing mode: all queues and exchanges will be temporary
    """
    connection_params = connection_params or {}
    if isinstance(connection_params, dict):
        connection = await connection_factory(**connection_params)
    else:
        connection = await connection_factory(connection_params)

    communicator = RmqCommunicator(
        connection=connection,

        # Messages
        message_exchange=message_exchange,
        queue_expires=queue_expires,

        # Tasks
        task_exchange=task_exchange,
        task_queue=task_queue,
        task_prefetch_size=task_prefetch_size,
        task_prefetch_count=task_prefetch_count,
        encoder=encoder,
        decoder=decoder,
        testing_mode=testing_mode
    )

    await communicator.connect()
    return communicator



# -*- coding: utf-8 -*-
import threading

import kiwipy


def on_broadcast_send(_comm, body, sender, subject, __):
    print(' [x] listening on_broadcast_send:')
    print(f' body: {body}, sender {sender}, subject {subject}\n')


def on_broadcast_filter(_comm, body, sender=None, subject=None, __=None):
    print(' [x] listening on_broadcast_filter:')
    print(f' body: {body}, sender {sender}, subject {subject}\n')


if __name__ == '__main__':
    filtered = kiwipy.BroadcastFilter(on_broadcast_filter)  # pylint: disable=invalid-name
    filtered.add_subject_filter('purchase.*')

    try:
        with kiwipy.connect('amqp://127.0.0.1') as comm:
            # Register a broadcast subscriber
            comm.add_broadcast_subscriber(on_broadcast_send)
            # Register a broadcast subscriber
            comm.add_broadcast_subscriber(filtered)
            # Now wait indefinitely for fibonacci calls
            threading.Event().wait()
    except KeyboardInterrupt:
        pass



# -*- coding: utf-8 -*-
import asyncio
from contextlib import contextmanager
import concurrent.futures
from concurrent.futures import Future as ThreadFuture
import functools
import logging
from typing import Union, Dict

import aio_pika
import deprecation
import pamqp
from pytray import aiothreads

import kiwipy
from kiwipy.version import __version__
from . import defaults
from . import communicator
from . import tasks

__all__ = 'RmqThreadCommunicator', 'RmqThreadTaskQueue', 'connect'

_LOGGER = logging.getLogger(__name__)


class RmqThreadCommunicator(kiwipy.Communicator):
    """
    RabbitMQ communicator that runs an event loop on a separate thread to do communication.
    This also means that heartbeats are not missed and the main program is free to block for
    as long as it wants.
    """

    TASK_TIMEOUT = 5.

    @classmethod
    def connect(
        cls,
        connection_params: Union[str, dict] = None,
        connection_factory=aio_pika.connect_robust,
        message_exchange=defaults.MESSAGE_EXCHANGE,
        task_exchange=defaults.TASK_EXCHANGE,
        task_queue=defaults.TASK_QUEUE,
        task_prefetch_size=defaults.TASK_PREFETCH_SIZE,
        task_prefetch_count=defaults.TASK_PREFETCH_COUNT,
        encoder=defaults.ENCODER,
        decoder=defaults.DECODER,
        testing_mode=False,
        async_task_timeout=TASK_TIMEOUT,
    ):
        # pylint: disable=too-many-arguments
        comm = cls(
            connection_params,
            connection_factory,
            message_exchange=message_exchange,
            task_exchange=task_exchange,
            task_queue=task_queue,
            task_prefetch_size=task_prefetch_size,
            task_prefetch_count=task_prefetch_count,
            encoder=encoder,
            decoder=decoder,
            testing_mode=testing_mode,
            async_task_timeout=async_task_timeout
        )

        # Start the communicator
        return comm

    def __init__(
        self,
        connection_params: Union[str, dict] = None,
        connection_factory=aio_pika.connect_robust,
        message_exchange: str = defaults.MESSAGE_EXCHANGE,
        queue_expires: int = defaults.QUEUE_EXPIRES,
        task_exchange=defaults.TASK_EXCHANGE,
        task_queue: str = defaults.TASK_QUEUE,
        task_prefetch_size=defaults.TASK_PREFETCH_SIZE,
        task_prefetch_count=defaults.TASK_PREFETCH_COUNT,
        encoder=defaults.ENCODER,
        decoder=defaults.DECODER,
        testing_mode=False,
        async_task_timeout=TASK_TIMEOUT,
    ):
        # pylint: disable=too-many-arguments
        """
        :param connection_params: parameters that will be passed to the connection factory to create the connection
        :param connection_factory: the factory method to open the aio_pika connection with
        :param message_exchange: The name of the RMQ message exchange to use
        :param queue_expires: the expiry time for standard queues in milliseconds. This is the time after which, if
            there are no subscribers, a queue will automatically be deleted by RabbitMQ.
        :param task_exchange: The name of the RMQ task exchange to use
        :param task_queue: The name of the task queue to use
        :param task_prefetch_count: the number of tasks this communicator can fetch simultaneously
        :param task_prefetch_size: the total size of the messages that the default queue can fetch simultaneously
        :param encoder: The encoder to call for encoding a message
        :param decoder: The decoder to call for decoding a message
        :param testing_mode: Run in testing mode: all queues and exchanges will be temporary
        """
        # Always use a separate loop
        self._loop = asyncio.new_event_loop()
        self._loop.set_debug(testing_mode)
        self._loop_scheduler = aiothreads.LoopScheduler(self._loop, 'RMQ communicator', async_task_timeout)
        self._stop_signal = None
        self._closed = False

        self._loop_scheduler.start()  # Star the loop scheduler (i.e. the event loop thread)

        # Establish the connection and get a communicator running on our thread
        self._communicator: communicator.RmqCommunicator = self._loop_scheduler.await_(
            communicator.async_connect(
                connection_params=connection_params,
                connection_factory=connection_factory,
                # Messages
                message_exchange=message_exchange,
                queue_expires=queue_expires,

                # Tasks
                task_exchange=task_exchange,
                task_queue=task_queue,
                task_prefetch_size=task_prefetch_size,
                task_prefetch_count=task_prefetch_count,
                encoder=encoder,
                decoder=decoder,
                testing_mode=testing_mode
            )
        )

    @deprecation.deprecated(
        deprecated_in='0.6.0',
        removed_in='0.7.0',
        current_version=__version__,
        details='This method was deprecated in v0.6.0 and will be removed in 0.7.0.  There is no '
        "replacement and the class should be reinstantiated if it's been closed."
    )
    def start(self):
        # pylint: disable=no-self-use
        raise DeprecationWarning(
            'This method was deprecated in v0.6.0 and will be removed in 0.7.0.  There is no '
            "replacement and the class should be reinstantiated if it's been closed."
        )

    @deprecation.deprecated(
        deprecated_in='0.6.0',
        removed_in='0.7.0',
        current_version=__version__,
        details='Use close() instead.  Unlike stop(), close() is permanent and '
        'the class should not be used again after this.'
    )
    def stop(self):
        self.close()

    @property
    def server_properties(self) -> Dict:
        """
        A dictionary containing server properties as returned by the RMQ server at connection time.

        The details are defined by the RMQ standard and can be found here:

        https://www.rabbitmq.com/amqp-0-9-1-reference.html#connection.start.server-properties

        The protocol states that this dictionary SHOULD contain at least:
            'host'          - specifying the server host name or address
            'product'       - giving the name of the server product
            'version'       - giving the name of the server version
            'platform'      - giving the name of the operating system
            'copyright'     - if appropriate, and,
            'information'   - giving other general information

        .. note:: In testing it seems like 'host' is not always returned.  Host information may be found in
            'cluster_name' but clients shouldn't rely on this.

        :return: the server properties dictionary
        """
        return self._communicator.server_properties

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def loop(self):
        return self._loop_scheduler.loop()

    def is_closed(self) -> bool:
        return self._closed

    def close(self):
        """Close down the communicator and clean up all resources.

        After this call it cannot be used again.
        """
        if self.is_closed():
            return

        self._loop_scheduler.await_(self._communicator.disconnect())
        self._loop_scheduler.close()
        self._loop.close()

        # Clean up
        del self._communicator
        del self._loop_scheduler
        del self._loop
        self._closed = True

    def add_close_callback(self, callback: aio_pika.types.CloseCallbackType, weak: bool = False) -> None:
        """Add a callable to be called each time (after) the connection is closed.

        :param weak: If True, the callback will be added to a `WeakSet`
        """
        self._ensure_open()
        self._communicator.add_close_callback(callback, weak)

    def add_rpc_subscriber(self, subscriber, identifier=None):
        self._ensure_open()
        return self._loop_scheduler.await_(
            self._communicator.add_rpc_subscriber(self._wrap_subscriber(subscriber), identifier)
        )

    def remove_rpc_subscriber(self, identifier):
        self._ensure_open()
        return self._loop_scheduler.await_(self._communicator.remove_rpc_subscriber(identifier))

    def add_task_subscriber(self, subscriber, identifier=None):
        self._ensure_open()
        return self._loop_scheduler.await_(
            self._communicator.add_task_subscriber(self._wrap_subscriber(subscriber), identifier)
        )

    def remove_task_subscriber(self, identifier):
        self._ensure_open()
        return self._loop_scheduler.await_(self._communicator.remove_task_subscriber(identifier))

    def add_broadcast_subscriber(self, subscriber, identifier=None):
        self._ensure_open()
        return self._loop_scheduler.await_(self._communicator.add_broadcast_subscriber(subscriber, identifier))

    def remove_broadcast_subscriber(self, identifier):
        self._ensure_open()
        return self._loop_scheduler.await_(self._communicator.remove_broadcast_subscriber(identifier))

    def task_send(self, task, no_reply=False):
        self._ensure_open()
        return self._loop_scheduler.await_(self._communicator.task_send(task, no_reply))

    def task_queue(
        self, queue_name: str, prefetch_size=defaults.TASK_PREFETCH_SIZE, prefetch_count=defaults.TASK_PREFETCH_COUNT
    ):
        self._ensure_open()
        aioqueue = self._loop_scheduler.await_(self._communicator.task_queue(queue_name, prefetch_size, prefetch_count))
        return RmqThreadTaskQueue(aioqueue, self._loop_scheduler, self._wrap_subscriber)

    def rpc_send(self, recipient_id, msg):
        self._ensure_open()
        return self._loop_scheduler.await_(self._communicator.rpc_send(recipient_id, msg))

    def broadcast_send(self, body, sender=None, subject=None, correlation_id=None):
        self._ensure_open()
        result = self._loop_scheduler.await_(
            self._communicator.broadcast_send(body=body, sender=sender, subject=subject, correlation_id=correlation_id)
        )
        return isinstance(result, pamqp.specification.Basic.Ack)

    def _wrap_subscriber(self, subscriber):
        """"
        We need to convert any kiwipy.Futures we get from a subscriber call into asyncio ones for
        the event loop based communicator.  Do this by wrapping any subscriber methods and
        intercepting the return values.
        """

        @functools.wraps(subscriber)
        def wrapper(*args, **kwargs):
            result = subscriber(*args, **kwargs)
            if isinstance(result, ThreadFuture):
                result = self._wrap_future(result)
            return result

        return wrapper

    def _wrap_future(self, kiwi_future: kiwipy.Future):
        aio_future = self._loop.create_future()

        def done(_):
            try:
                result = kiwi_future.result()
            except concurrent.futures.CancelledError:
                self._loop.call_soon_threadsafe(aio_future.cancel)
            except Exception as exc:  # pylint: disable=broad-except
                self._loop.call_soon_threadsafe(aio_future.set_exception, exc)
            else:
                if isinstance(result, kiwipy.Future):
                    result = self._wrap_future(result)
                self._loop.call_soon_threadsafe(aio_future.set_result, result)

        kiwi_future.add_done_callback(done)
        return aio_future

    def _ensure_open(self):
        if self.is_closed():
            raise kiwipy.CommunicatorClosed


class RmqThreadTaskQueue:
    """
    Thread task queue.
    """

    def __init__(self, task_queue: tasks.RmqTaskQueue, loop_scheduler: aiothreads.LoopScheduler, wrap_subscriber):
        self._task_queue = task_queue
        self._loop_scheduler = loop_scheduler
        self._wrap_subscriber = wrap_subscriber

    def __iter__(self):
        for task in self._loop_scheduler.async_iter(self._task_queue):
            yield RmqThreadIncomingTask(task, self._loop_scheduler)

    def task_send(self, task, no_reply=False):
        return self._loop_scheduler.await_(self._task_queue.task_send(task, no_reply))

    def add_task_subscriber(self, subscriber):
        return self._loop_scheduler.await_(self._task_queue.add_task_subscriber(self._wrap_subscriber(subscriber)))

    def remove_task_subscriber(self, subscriber):
        # Note: This probably doesn't work as in add_task_subscriber we wrap it and so
        # it will be a different function here
        return self._loop_scheduler.await_(self._task_queue.remove_task_subscriber(subscriber))

    @contextmanager
    def next_task(self, timeout=5., fail=True):
        with self._loop_scheduler.async_ctx(self._task_queue.next_task(timeout=timeout, fail=fail)) as task:
            yield RmqThreadIncomingTask(task, self._loop_scheduler)


class RmqThreadIncomingTask:

    def __init__(self, task: tasks.RmqIncomingTask, loop_scheduler):
        self._task = task
        self._loop_scheduler = loop_scheduler

    @property
    def body(self) -> str:
        return self._task.body

    @property
    def no_reply(self) -> bool:
        return self._task.no_reply

    @property
    def state(self) -> str:
        return self._task.state

    def process(self) -> ThreadFuture:
        return aiothreads.aio_future_to_thread(self._task.process())

    def requeue(self):
        self._task.requeue()

    @contextmanager
    def processing(self):
        with self._loop_scheduler.ctx(self._task.processing()) as outcome:
            yield outcome


def connect(
    connection_params: Union[str, dict] = None,
    connection_factory=aio_pika.connect_robust,
    message_exchange=defaults.MESSAGE_EXCHANGE,
    task_exchange=defaults.TASK_EXCHANGE,
    task_queue=defaults.TASK_QUEUE,
    task_prefetch_size=defaults.TASK_PREFETCH_SIZE,
    task_prefetch_count=defaults.TASK_PREFETCH_COUNT,
    encoder=defaults.ENCODER,
    decoder=defaults.DECODER,
    testing_mode=False,
) -> RmqThreadCommunicator:
    """
    Establish a RabbitMQ communicator connection
    """
    # pylint: disable=too-many-arguments
    _communicator = RmqThreadCommunicator.connect(
        connection_params=connection_params,
        connection_factory=connection_factory,
        message_exchange=message_exchange,
        task_exchange=task_exchange,
        task_queue=task_queue,
        task_prefetch_size=task_prefetch_size,
        task_prefetch_count=task_prefetch_count,
        encoder=encoder,
        decoder=decoder,
        testing_mode=testing_mode
    )
    return _communicator



# -*- coding: utf-8 -*-
import abc
import sys
from typing import Any, Callable

import shortuuid

from . import exceptions
# For backwards compatibility import exceptions too
from .exceptions import *  # pylint: disable=wildcard-import, redefined-builtin, unused-wildcard-import
from . import futures

__all__ = 'Communicator', 'CommunicatorHelper'

# RPC subscriber params: communicator, msg
RpcSubscriber = Callable[['Communicator', Any], Any]
# Task subscriber params: communicator, task
TaskSubscriber = Callable[['Communicator', Any], Any]
# Broadcast subscribers params: communicator, body, sender, subject, correlation id
BroadcastSubscriber = Callable[['Communicator', Any, Any, Any, Any], Any]


class Communicator:
    """
    The interface for a communicator used to both send and receive various types of message.
    """

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @abc.abstractmethod
    def is_closed(self) -> bool:
        """Return `True` if the communicator was closed"""

    @abc.abstractmethod
    def close(self):
        """Close a communicator, free up all resources and do not allow any further operations"""

    @abc.abstractmethod
    def add_rpc_subscriber(self, subscriber: RpcSubscriber, identifier=None) -> Any:
        """Add an RPC subscriber to the communicator with an optional identifier.  If an identifier
        is not provided the communicator will generate a unique one.  In all cases the identifier
        will be returned."""

    @abc.abstractmethod
    def remove_rpc_subscriber(self, identifier):
        """
        Remove an RPC subscriber given the identifier.  Raises a `ValueError` if there
        is no such subscriber.

        :param identifier: The RPC subscriber identifier
        """

    @abc.abstractmethod
    def add_task_subscriber(self, subscriber: TaskSubscriber, identifier=None) -> Any:
        """Add a task subscriber to the communicator's default queue.  Returns the identifier.

        :param subscriber: The task callback function
        :param identifier: the subscriber identifier
        """

    @abc.abstractmethod
    def remove_task_subscriber(self, identifier):
        """Remove a task subscriber from the communicator's default queue.

        :param identifier: the subscriber to remove
        :raises: ValueError if identifier does not correspond to a known subscriber
        """

    @abc.abstractmethod
    def add_broadcast_subscriber(self, subscriber: BroadcastSubscriber, identifier=None) -> Any:
        """
        Add a broadcast subscriber that will receive all broadcast messages

        :param subscriber: the subscriber function to be called
        :param identifier: an optional identifier for the subscriber
        :return: an identifier for the subscriber and can be subsequently used to remove it
        """

    @abc.abstractmethod
    def remove_broadcast_subscriber(self, identifier):
        """
        Remove a broadcast subscriber

        :param identifier: the identifier of the subscriber to remove
        """

    @abc.abstractmethod
    def task_send(self, task, no_reply=False) -> futures.Future:
        """
        Send a task messages, this will be queued and picked up by a
        worker at some point in the future.  The method returns a future
        representing the outcome of the task.

        :param task: The task message
        :param no_reply: Do not send a reply containing the result of the task
        :type no_reply: bool
        :return: A future corresponding to the outcome of the task
        """

    @abc.abstractmethod
    def rpc_send(self, recipient_id, msg):
        """
        Initiate a remote procedure call on a recipient.  This method
        returns a future representing the outcome of the call.

        :param recipient_id: The recipient identifier
        :param msg: The body of the message
        :return: A future corresponding to the outcome of the call
        :rtype: :class:`kiwipy.Future`
        """

    @abc.abstractmethod
    def broadcast_send(self, body, sender=None, subject=None, correlation_id=None) -> bool:
        """Broadcast a message to all subscribers"""


class CommunicatorHelper(Communicator):
    # Have to disable this linter because this class remains abstract and it is
    # just used by classes that will themselves be concrete
    # pylint: disable=abstract-method

    def __init__(self):
        self._task_subscribers = {}
        self._broadcast_subscribers = {}
        self._rpc_subscribers = {}
        self._closed = False

    def is_closed(self) -> bool:
        return self._closed

    def close(self):
        if self._closed:
            return
        self._closed = True
        del self._task_subscribers
        del self._broadcast_subscribers
        del self._rpc_subscribers

    def add_rpc_subscriber(self, subscriber, identifier=None) -> Any:
        self._ensure_open()
        identifier = identifier or shortuuid.uuid()
        if identifier in self._rpc_subscribers:
            raise exceptions.DuplicateSubscriberIdentifier(f"RPC identifier '{identifier}'")
        self._rpc_subscribers[identifier] = subscriber
        return identifier

    def remove_rpc_subscriber(self, identifier):
        self._ensure_open()
        try:
            self._rpc_subscribers.pop(identifier)
        except KeyError:
            raise ValueError(f"Unknown subscriber '{identifier}'")

    def add_task_subscriber(self, subscriber, identifier=None):
        """
        Register a task subscriber

        :param subscriber: The task callback function
        :param identifier: the subscriber identifier
        """
        self._ensure_open()
        identifier = identifier or shortuuid.uuid()
        if identifier in self._rpc_subscribers:
            raise exceptions.DuplicateSubscriberIdentifier(f"RPC identifier '{identifier}'")
        self._task_subscribers[identifier] = subscriber
        return identifier

    def remove_task_subscriber(self, identifier):
        """
        Remove a task subscriber

        :param identifier: the subscriber to remove
        :raises: ValueError if identifier does not correspond to a known subscriber
        """
        self._ensure_open()
        try:
            self._task_subscribers.pop(identifier)
        except KeyError:
            raise ValueError(f"Unknown subscriber: '{identifier}'")

    def add_broadcast_subscriber(self, subscriber: BroadcastSubscriber, identifier=None) -> Any:
        self._ensure_open()
        identifier = identifier or shortuuid.uuid()
        if identifier in self._broadcast_subscribers:
            raise exceptions.DuplicateSubscriberIdentifier(f"Broadcast identifier '{identifier}'")

        self._broadcast_subscribers[identifier] = subscriber
        return identifier

    def remove_broadcast_subscriber(self, identifier):
        self._ensure_open()
        try:
            del self._broadcast_subscribers[identifier]
        except KeyError:
            raise ValueError(f"Broadcast subscriber '{identifier}' unknown")

    def fire_task(self, msg, no_reply=False):
        self._ensure_open()
        future = futures.Future()

        for subscriber in self._task_subscribers.values():
            try:
                result = subscriber(self, msg)
                future.set_result(result)
                break
            except exceptions.TaskRejected:
                pass
            except Exception:  # pylint: disable=broad-except
                future.set_exception(exceptions.RemoteException(sys.exc_info()))
                break

        if no_reply:
            return None

        return future

    def fire_rpc(self, recipient_id, msg):
        self._ensure_open()
        try:
            subscriber = self._rpc_subscribers[recipient_id]
        except KeyError:
            raise exceptions.UnroutableError(f"Unknown rpc recipient '{recipient_id}'")
        else:
            future = futures.Future()
            try:
                future.set_result(subscriber(self, msg))
            except Exception:  # pylint: disable=broad-except
                future.set_exception(exceptions.RemoteException(sys.exc_info()))

            return future

    def fire_broadcast(self, body, sender=None, subject=None, correlation_id=None):
        self._ensure_open()
        for subscriber in self._broadcast_subscribers.values():
            subscriber(self, body=body, sender=sender, subject=subject, correlation_id=correlation_id)
        return True

    def _ensure_open(self):
        if self.is_closed():
            raise exceptions.CommunicatorClosed



# -*- coding: utf-8 -*-
import kiwipy

# pylint: disable=invalid-name

with kiwipy.connect('amqp://127.0.0.1') as comm:
    # Send an RPC message with identifier 'fib'
    print(' [x] Requesting fib(30)')
    response = comm.rpc_send('fib', 30).result()
    print(f' [.] Got {response!r}')

    # Send an RPC message with identifier 'fac'
    print(' [x] Requesting fac(3)')
    response = comm.rpc_send('fac', 3).result()
    print(f' [.] Got {response!r}')



# -*- coding: utf-8 -*-
import logging
import traceback

__all__ = ('EventHelper',)

_LOGGER = logging.getLogger(__name__)


class EventHelper:

    def __init__(self, listener_type):
        assert listener_type is not None, 'Must provide valid listener type'

        self._listener_type = listener_type
        self._listeners = set()

    def add_listener(self, listener):
        assert isinstance(listener, self._listener_type), 'Listener is not of right type'
        self._listeners.add(listener)

    def remove_listener(self, listener):
        self._listeners.discard(listener)

    def remove_all_listeners(self):
        self._listeners.clear()

    @property
    def listeners(self):
        return self._listeners

    def fire_event(self, event_function, *args, **kwargs):
        if event_function is None:
            raise ValueError('Must provide valid event method')

        for listener in self.listeners:
            try:
                getattr(listener, event_function.__name__)(*args, **kwargs)
            except Exception:  # pylint: disable=broad-except
                _LOGGER.error('Listener %s produced an exception:\n%s', listener, traceback.format_exc())



# -*- coding: utf-8 -*-
from . import communications

__all__ = ('LocalCommunicator',)


class LocalCommunicator(communications.CommunicatorHelper):

    def task_send(self, task, no_reply=False):
        return self.fire_task(task, no_reply)

    def rpc_send(self, recipient_id, msg):
        return self.fire_rpc(recipient_id, msg)

    def broadcast_send(self, body, sender=None, subject=None, correlation_id=None):
        return self.fire_broadcast(body, sender=sender, subject=subject, correlation_id=correlation_id)



# -*- coding: utf-8 -*-
import sys

import kiwipy

# pylint: disable=invalid-name

body = ' '.join(sys.argv[1:]) or '___'

with kiwipy.connect('amqp://127.0.0.1') as comm:
    # send message with different sender and subject

    # listen by two subscriber
    sendr = 'bob.jones'
    subj = 'purchase.car'
    comm.broadcast_send(body, sender=sendr, subject=subj)

    # Filtered by filter subscriber because subject not matched with "purchase.*" pattern
    # Therefore filterd.
    sendr = 'bob.jones'
    subj = 'sell.car'
    comm.broadcast_send(body, sender=sendr, subject=subj)



# -*- coding: utf-8 -*-
import asyncio
import collections.abc
import functools
import inspect
import os
import socket
import traceback

from kiwipy import exceptions

__all__ = ()

# The key used in messages to give information about the host that send a message
HOST_KEY = 'host'
HOSTNAME_KEY = 'hostname'
PID_KEY = 'pid'
RESULT_KEY = 'result'
EXCEPTION_KEY = 'exception'
CANCELLED_KEY = 'cancelled'
PENDING_KEY = 'pending'


def get_host_info():
    return {'hostname': socket.gethostname(), 'pid': os.getpid()}


def add_host_info(msg):
    if HOST_KEY in msg:
        raise ValueError('Host information key already exists in message')

    msg[HOST_KEY] = get_host_info()


def result_response(result):
    return {RESULT_KEY: result}


def exception_response(exception: Exception, trace=None) -> dict:
    """
    Create an exception response dictionary
    :param exception: The exception to encode
    :param trace: Optional traceback
    :return: An exception response dictionary
    """
    msg = str(exception)
    if trace is not None:
        msg += f"\n{''.join(traceback.format_tb(trace)[0])}"
    return {EXCEPTION_KEY: msg}


def cancelled_response(msg=None):
    return {CANCELLED_KEY: msg}


def pending_response(msg=None):
    return {PENDING_KEY: msg}


def response_result(response):
    future = asyncio.Future()
    response_to_future(response, future)
    return future.result()


def response_to_future(response, future=None):
    """
    Take a response message and set the appropriate value on the given future
    :param response:
    :param future:
    :return:
    """
    if not isinstance(response, collections.abc.Mapping):
        raise TypeError('Response must be a mapping')

    if future is None:
        future = asyncio.Future()

    if CANCELLED_KEY in response:
        future.cancel()
    elif EXCEPTION_KEY in response:
        future.set_exception(exceptions.RemoteException(response[EXCEPTION_KEY]))
    elif RESULT_KEY in response:
        future.set_result(response[RESULT_KEY])
    elif PENDING_KEY in response:
        future.set_result(asyncio.Future())
    else:
        raise ValueError(f"Unknown response type '{response}'")

    return future


def future_to_response(future):
    """
    Convert a future to a response dictionary
    :param future: The future
    :type future: :class:`kiwipy.Future`
    :return: The response dictionary
    :rtype: dict
    """
    if future.cancelled():
        return cancelled_response()
    try:
        return result_response(future.result())
    except Exception as exception:  # pylint: disable=broad-except
        return exception_response(exception)


def ensure_coroutine(coro_or_fn):
    if asyncio.iscoroutinefunction(coro_or_fn):
        return coro_or_fn
    if callable(coro_or_fn):
        if inspect.isclass(coro_or_fn):
            coro_or_fn = coro_or_fn.__call__

        @functools.wraps(coro_or_fn)
        async def wrap(*args, **kwargs):
            return coro_or_fn(*args, **kwargs)

        return wrap

    raise TypeError('coro_or_fn must be a callable')



# -*- coding: utf-8 -*-
from setuptools import setup

__author__ = 'Martin Uhrin'
__license__ = 'GPLv3 and MIT'
__contributors__ = ['Sebastiaan P. Huber', 'Jason Yu', 'Sonia Collaud']

ABOUT = {}
with open('kiwipy/version.py') as f:
    exec(f.read(), ABOUT)  # pylint: disable=exec-used

setup(
    name='kiwipy',
    version=ABOUT['__version__'],
    description='Robust, high-volume, message based communication made easy',
    long_description=open('README.rst').read(),
    url='https://github.com/aiidateam/kiwipy.git',
    author='Martin Uhrin',
    author_email='martin.uhrin.10@ucl.ac.uk',
    license=__license__,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    keywords='communication messaging rpc broadcast',
    install_requires=['shortuuid', 'async_generator', 'pytray>=0.2.2, <0.4.0', 'deprecation'],
    python_requires='>=3.7',
    extras_require={
        'docs': [
            'docutils==0.14',
            'jupyter',  # For running doc examples
            'nbsphinx',  # Jupyter notebooks in docs
            'pandoc',
            'sphinx',
            'sphinx-autobuild',
        ],
        'pre-commit': ['pre-commit~=2.2', 'pylint==2.5.2'],
        'rmq': ['aio-pika', 'pyyaml~=5.1'],
        'tests': [
            'coverage',
            'ipykernel',
            'pytest-cov',
            'pytest~=5.4',
            'pytest-asyncio~=0.12,<0.17',
            'pytest-notebook>=0.7',
            'pytest-benchmark',
            'pika',
            'msgpack',
        ]
    },
    packages=['kiwipy', 'kiwipy.rmq'],
    test_suite='test'
)



# -*- coding: utf-8 -*-
import time
import threading

import kiwipy

print(' [*] Waiting for messages. To exit press CTRL+C')


def callback(_comm, task):
    print(f' [x] Received {task!r}')
    time.sleep(task.count('.'))
    print(' [x] Done')
    return task


with kiwipy.connect('amqp://127.0.0.1/') as communicator:
    communicator.add_task_subscriber(callback)
    try:
        threading.Event().wait()
    except KeyboardInterrupt:
        pass



# -*- coding: utf-8 -*-
import threading

import kiwipy


def fib(_comm, num):
    if num == 0:
        return 0
    if num == 1:
        return 1
    return fib(_comm, num - 1) + fib(_comm, num - 2)


def fac(_comm, num):
    result = 1
    if num > 1:
        result = num * fac(_comm, num - 1)
    return result


if __name__ == '__main__':
    try:
        with kiwipy.connect('amqp://127.0.0.1') as comm:
            # Register an RPC subscriber with the name 'fib'
            comm.add_rpc_subscriber(fib, 'fib')
            # Register an RPC subscriber with the name 'fac'
            comm.add_rpc_subscriber(fac, 'fac')
            # Now wait indefinitely for fibonacci calls
            threading.Event().wait()
    except KeyboardInterrupt:
        pass
